import os
import sys
import random
from functools import lru_cache
from abc import ABC, abstractmethod

import torch
from torch import nn
import numpy as np
from mpi4py import MPI
import gym
from pipcs import Config


@lru_cache(maxsize=1)
def _center_function(population_size):
    centers = np.arange(0, population_size, dtype=np.float32)
    centers = centers / (population_size - 1)
    centers -= 0.5
    return centers


def _compute_ranks(rewards):
    rewards = np.array(rewards)
    ranks = np.empty(rewards.size, dtype=int)
    ranks[rewards.argsort()] = np.arange(rewards.size)
    return ranks


def rank_transformation(rewards):
    ranks = _compute_ranks(rewards)
    values = _center_function(rewards.size)
    return values[ranks]


class Policy(nn.Module, ABC):
    """Abstract subclass of nn.Module."""
    def __init__(self):
        super().__init__()

    @abstractmethod
    def rollout(self, env):
        """This function should be implemented by the user and
        it should evaluate the model in the given environment and
        return the total reward."""
        pass


class NES():
    """
    :ivar int gen: Current generation
    :ivar Policy policy: Trained policy
    :ivar torch.optim.Optimizer optim: Optimizer of the policy
    """
    def __init__(self, config: Config):
        config.check_config()
        self.config = config

        torch.manual_seed(config.nes.seed)
        np.random.seed(config.nes.seed)
        random.seed(config.nes.seed)
        torch.cuda.manual_seed(config.nes.seed)
        torch.cuda.manual_seed_all(config.nes.seed)
        torch.backends.cudnn.benchmark = False
        torch.backends.cudnn.deterministic = True
        self.env = self.make_env(**config.environment.to_dict())
        self.env.seed(config.nes.seed)
        self.env.action_space.seed(config.nes.seed)

        self.gen = 0

        self.policy = self.make_policy(**config.policy.to_dict())
        self.optim = self.make_optimizer(policy=self.policy, **config.optimizer.to_dict())

    @staticmethod
    def make_env(make_env, **kwargs):
        """Helper function to create a gym environment."""
        return make_env(**kwargs)

    @staticmethod
    def make_policy(policy, device, **kwargs):
        """Helper function to create a policy."""
        assert issubclass(policy, Policy)
        return policy(**kwargs).to(device)

    @staticmethod
    def make_optimizer(policy, optim_type, **kwargs):
        """Helper function to create an optimizer."""
        return optim_type(policy.parameters(), **kwargs)

    def eval_policy(self, policy):
        """Evaluate policy on the ``self.env`` for ``self.config.nes.n_rollout times``"""
        total_reward = 0
        for _ in range(self.config.nes.n_rollout):
            total_reward += policy.rollout(self.env)
        return total_reward / self.config.nes.n_rollout

    def train(self):
        """Train ``self.policy`` for ``self.config.nes.n_steps`` to increase reward returns
        from the ``self.env`` using Natural Evolution Strategy gradient estimation."""
        torch.set_grad_enabled(False)
        comm = MPI.COMM_WORLD
        n_workers = comm.Get_size()
        rank = comm.Get_rank()
        if not rank == 0:
            f = open(os.devnull, 'w')
            sys.stdout = f
        device = self.config.policy.device
        dummy_policy = self.make_policy(**self.config.policy.to_dict())
        for gen in range(self.config.nes.n_step):
            self.gen = gen

            # Sample
            mean = torch.nn.utils.parameters_to_vector(self.policy.parameters())
            normal = torch.distributions.normal.Normal(0, self.config.nes.sigma)
            epsilon = normal.sample([int(self.config.nes.population_size/2), mean.shape[0]])
            population_params = torch.cat((mean + epsilon, mean - epsilon))
            epsilons = torch.cat((epsilon, -epsilon))

            # Evaluate
            rewards = []
            reward_array = np.zeros(self.config.nes.population_size, dtype=np.float32)
            batch = np.array_split(population_params, n_workers)[rank]
            for param in batch:
                torch.nn.utils.vector_to_parameters(param.to(device), dummy_policy.parameters())
                rewards.append(self.eval_policy(dummy_policy))
            comm.Allgatherv([np.array(rewards, dtype=np.float32), MPI.FLOAT], [reward_array, MPI.FLOAT])

            # Calculate gradients
            ranked_rewards = torch.from_numpy(rank_transformation(reward_array)).unsqueeze(0).to(device)
            grad = -(torch.mm(ranked_rewards, epsilons) / (len(reward_array) * self.config.nes.sigma))
            grad = (grad + mean * self.config.nes.l2_decay).squeeze()  # L2 Decay

            # Optimize
            self.optim.zero_grad()
            index = 0
            for parameter in self.policy.parameters():
                size = np.prod(parameter.shape)
                parameter.grad = grad[index:index+size].view(parameter.shape).to(device)
                # Limit gradient update to increase stability.
                parameter.grad.data.clamp_(-1.0, 1.0)
                index += size
            self.optim.step()

            self.config.nes.after_optimize_hook(self)

        sys.stdout = sys.__stdout__
        torch.set_grad_enabled(True)



import hashlib
import cloudpickle


class Logger():
    def __init__(self, config):
        # check config.log.path


        # if os.path.exists(path) and os.path.isfile(path):
        #     self.file = open(path, 'a')
        # else:
        #     if not os.path.exists(os.path.dirname(path))
        #         os.makedirs(os.path.dirname(path))

        # assert os.path.isfile(path)

    def __exit__(self):





from typing import Callable, List, Optional
from dataclasses import field
from itertools import tee

import gym
from pipcs import Config
import torch
from torch import nn
import numpy as np

from nes import NES, Policy, default_config


def pairwise(iterable):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


class Agent(Policy):
    def __init__(self, input_size, hidden_layers, output_size, output_func,
                 hidden_act, output_act=None, bias=True):
        super().__init__()
        layers: List[nn.Module] = []
        layer_sizes = [input_size] + hidden_layers + [output_size]
        for idx, (layer1, layer2) in enumerate(pairwise(layer_sizes)):
            layers.append(nn.Linear(layer1, layer2, bias=bias))
            if idx < len(layer_sizes) - 2:
                layers.append(hidden_act())
        if output_act is not None:
            layers.append(output_act())
        self.output_func = output_func
        self.seq = nn.Sequential(*layers)

    def rollout(self, env):
        done = False
        obs = env.reset()
        total_reward = 0
        while not done:
            obs_tensor = torch.from_numpy(obs).float()
            action = self.output_func(self.seq(obs_tensor))
            obs, reward, done, _ = env.step(action)
            total_reward += reward
        return total_reward


def after_optimize_hook(self):
    reward = self.eval_policy(self.policy)
    if reward > self.config.user_variables.best_reward:
        self.config.user_variables.best_reward = reward
    best_reward = self.config.user_variables.best_reward
    print(f'Gen: {self.gen} Test Reward: {reward} Best Reward: {best_reward}')


config = Config(default_config)


def make_env(id):
    return gym.make(id)


@config('environment')
class EnvironmentConfig():
    make_env = make_env
    id: str = 'LunarLander-v2'


@config('policy')
class PolicyConfig():
    policy = Agent
    env = NES.make_env(**config.environment.to_dict())
    input_size: int = env.observation_space.shape[0]
    if isinstance(env.action_space, gym.spaces.Discrete):
        output_size: int = env.action_space.n
        output_func: Callable[[int], int] = lambda x: x.argmax().item()
    else:
        output_size: int = env.action_space.shape[0]
        output_func: Callable[[int], int] = lambda x: x.detach().numpy()
    hidden_layers: List[int] = field(default_factory=lambda: [64, 32])
    hidden_act: nn.Module = nn.ReLU
    output_act: Optional[nn.Module] = nn.Tanh
    bias: bool = True


@config('optimizer')
class OptimizerConfig():
    lr = 0.02
    optim_type = torch.optim.Adam


@config('nes')
class NESConfig():
    n_rollout = 5
    n_step = 500
    l2_decay = 0.005
    population_size = 256
    sigma = 0.02
    seed = 123123
    after_optimize_hook = after_optimize_hook


@config('user_variables')
class UserVariables():
    """Users can initialize variables as config to use in the hook function."""
    best_reward: float = -np.inf


if __name__ == '__main__':
    nes = NES(config)
    nes.train()

    # Evaluate trained policy
    print(f'Reward: {nes.eval_policy(nes.policy)}')
